// Dir2File.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <windows.h>
#include <assert.h>
#include <conio.h>
#include <stdlib.h>

#define UNICODE_SIGNATURE WORD(0xFEFF)

/* Calculating file hash value */
int calcHash(LPCTSTR file_path){
	// create handle for the file
	HANDLE hFile=CreateFile(file_path,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_FLAG_RANDOM_ACCESS,NULL);
	assert(hFile!=INVALID_HANDLE_VALUE);

	/* compute hash value */
	BYTE bFirst, bLast;
	DWORD dwBytesRead;
	//	read first byte from file
	BOOL bRes=ReadFile(hFile,&bFirst,sizeof(BYTE),&dwBytesRead,NULL);
	assert(bRes && (sizeof(BYTE)==dwBytesRead));
	//move file pointer to last byte of file
	DWORD dwSetPointer = SetFilePointer(hFile,-1,NULL,FILE_END);
	assert(dwSetPointer!=INVALID_SET_FILE_POINTER);
	// read last byte from file
	bRes=ReadFile(hFile,&bLast,sizeof(BYTE),&dwBytesRead,NULL);
	assert(bRes && (sizeof(BYTE)==dwBytesRead));
	// close file handle
	CloseHandle(hFile);

	// compute hash value
	WORD wValue = bFirst + bLast;
	wValue /= 2;

	return wValue;
}

/* calculating input folder disk cluster size in bytes */
DWORD calcDiskClusterSize(LPCTSTR srcInputFolder){
	// input folder root directory
	TCHAR root[MAX_PATH];
	_tsplitpath(srcInputFolder,root,NULL,NULL,NULL);

	// compute disk cluster size
	DWORD dwBytesPerSector, dwSectorsPerCluster;
	BOOL bRes=GetDiskFreeSpace(root,&dwSectorsPerCluster,&dwBytesPerSector,NULL,NULL);
	assert(bRes);

	// return cluster size in bytes
	return (dwBytesPerSector*dwSectorsPerCluster);
}

/* get input folder and output file path; list input folder files and files-info in output file */
int _tmain(int argc, _TCHAR* argv[])
{
	// input check
	assert(argc==3);
	LPCTSTR srcInputFolder=argv[1];	// input folder
	LPCTSTR srcOutputPath=argv[2];	// output path

	// create\override output file
	TCHAR output_path[MAX_PATH];
	_tcscpy_s(output_path, srcOutputPath);
	
	HANDLE hOutputFile=CreateFile(output_path,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_FLAG_SEQUENTIAL_SCAN,NULL);
	assert(hOutputFile!=INVALID_HANDLE_VALUE);

	// write unicode signature to output file
	WORD wUnicodeSign = UNICODE_SIGNATURE;
	DWORD dwBytesWritten;
	BOOL bRes=WriteFile(hOutputFile,&wUnicodeSign,sizeof(WORD),&dwBytesWritten,NULL);
	assert(bRes && (dwBytesWritten==sizeof(WORD)));

	// compute input folder disk cluster size
	DWORD dwClusterSize=calcDiskClusterSize(srcInputFolder);

	// start search
	TCHAR search_string[MAX_PATH];
	_tcscpy_s(search_string,srcInputFolder);
	_tcscat_s(search_string,_T("*.*"));	// search all files in source folder

	WIN32_FIND_DATA w32fd;
	ZeroMemory(&w32fd,sizeof(w32fd));
	
	HANDLE hFindFile=FindFirstFile(search_string,&w32fd);
	bRes=(hFindFile!=INVALID_HANDLE_VALUE);

	// while there are unread files in the input folder
	while(bRes)
	{
		// handle the file only if it's not a subfolder, dotted folder or a system file
		if (!(w32fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) && !(w32fd.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM)){
			// file name
			LPCTSTR fileName=w32fd.cFileName;

			// file size (up to 2MB)
			DWORD dwFileSize=w32fd.nFileSizeLow;

			// file size on disk; will be 0 if file size is 0
			DWORD dwFileSizeOnDisk=0;
			if (dwFileSize!=0){
				DWORD dwClustersPerFile=dwFileSize/dwClusterSize+(dwFileSize%dwClusterSize?1:0);
				dwFileSizeOnDisk=dwClustersPerFile*dwClusterSize;
			}

			// file hash value; will be 0 if file size is 0
			DWORD dwFileHashValue=0;
			if (dwFileSize!=0){
				TCHAR file_path[MAX_PATH];
				_tcscpy_s(file_path,srcInputFolder);
				_tcscat_s(file_path,fileName);
				dwFileHashValue=calcHash(file_path);
			}

			// write data to string
			TCHAR lineToOutput[_MAX_FNAME + _MAX_EXT + 30]; // should cover one line length
			_stprintf(lineToOutput,_T("%s\t%d\t%d\t%d\r\n"),fileName,dwFileSize,dwFileSizeOnDisk,dwFileHashValue);

			// write string to output file
			DWORD dwLineBytesSize=_tcslen(lineToOutput)*2; // 2 bytes per char
			bRes=WriteFile(hOutputFile,lineToOutput,dwLineBytesSize,&dwBytesWritten,NULL);
			assert(bRes && (dwBytesWritten==dwLineBytesSize));
		} // end if

		// get next file
		bRes=FindNextFile(hFindFile,&w32fd);

	} // end while

	// assert end of file list and not error getting next file in folder
	assert(GetLastError()==ERROR_NO_MORE_FILES);
	
	// close file handles and exit
	FindClose(hFindFile);
	CloseHandle(hOutputFile);

	return 0;
}