// reader.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <stdio.h>
#include <tchar.h>
#include <windows.h>
#include <assert.h>

#define QUEUE_MAX_COUNT			13
#define MMF_NAME				_T("EX3.SHARED.QUEUE.MMF")
#define GUARD_MUTEX_NAME		_T("EX3.SHARED.QUEUE.MUTEX")
#define ALMOST_INFINITE         100000
#define READER_SEMAPHORE		_T("EX3.READER_SEMAPHORE")
#define WRITER_SEMAPHORE		_T("EX3.WRITER_SEMAPHORE")

typedef struct
{
	DWORD m_dwCount;
	DWORD m_dwHead;
	DWORD m_dwTail;

} SHARED_QUEUE_HEADER;

typedef enum
{
	CC_PUSH,
	CC_POP,
	CC_EXIT
	
} COMMAND_CODE;

typedef struct
{
	COMMAND_CODE	m_cc;
	DWORD			m_dwValue;
} QUEUE_COMMAND;

typedef struct
{
	SHARED_QUEUE_HEADER* m_pHeader;
	DWORD*				 m_pQueue;
	HANDLE				 m_hMutex;

} SHARED_QUEUE;

BOOL InitializeSharedQueue(SHARED_QUEUE* pSQ)
{
	DWORD dwRes=WaitForSingleObject(pSQ->m_hMutex,ALMOST_INFINITE);
	assert(dwRes==WAIT_OBJECT_0);

	ZeroMemory(pSQ->m_pHeader,sizeof(SHARED_QUEUE_HEADER));

	ReleaseMutex(pSQ->m_hMutex);

	return TRUE;
}

BOOL PushToQueue(SHARED_QUEUE* pSQ,DWORD dwVal)
{
	if(QUEUE_MAX_COUNT==pSQ->m_pHeader->m_dwCount)
		return FALSE;

	pSQ->m_pQueue[pSQ->m_pHeader->m_dwHead++]=dwVal;

	if(pSQ->m_pHeader->m_dwHead==QUEUE_MAX_COUNT)
		pSQ->m_pHeader->m_dwHead=0;

	pSQ->m_pHeader->m_dwCount++;

	return TRUE;
}
BOOL PopFromQueue(SHARED_QUEUE* pSQ,DWORD* pVal)
{
	if(!pSQ->m_pHeader->m_dwCount)
		return FALSE;

	(*pVal)=pSQ->m_pQueue[pSQ->m_pHeader->m_dwTail++];

	if(pSQ->m_pHeader->m_dwTail==QUEUE_MAX_COUNT)
		pSQ->m_pHeader->m_dwTail=0;

	pSQ->m_pHeader->m_dwCount--;

	return TRUE;
}

BOOL ProcessQueueCommand(SHARED_QUEUE* pSQ,QUEUE_COMMAND* pCommand)
{
	DWORD dwRes=WaitForSingleObject(pSQ->m_hMutex,ALMOST_INFINITE);
	assert(dwRes==WAIT_OBJECT_0);
	BOOL bRes;

	if(pCommand->m_cc==CC_PUSH)
	{
		bRes=PushToQueue(pSQ,pCommand->m_dwValue);
		
	}
	else 
	{
		bRes=PopFromQueue(pSQ,&pCommand->m_dwValue);
	}

	BOOL bRes1=ReleaseMutex(pSQ->m_hMutex);
	assert(bRes1);

	return bRes;
}

int _tmain(int argc, _TCHAR* argv[])
{
	// creating shared queue in memory
	DWORD dwSize=QUEUE_MAX_COUNT* sizeof(DWORD)+sizeof(SHARED_QUEUE_HEADER);
	SHARED_QUEUE shared_queue;

	HANDLE hMapFile = CreateFileMapping(INVALID_HANDLE_VALUE,NULL,PAGE_READWRITE,0,dwSize,MMF_NAME);  
	assert(hMapFile);

	BOOL bConnect=(GetLastError()==ERROR_ALREADY_EXISTS);

	BYTE* pBuffer=(BYTE*)MapViewOfFile(hMapFile,FILE_MAP_WRITE|FILE_MAP_WRITE,0,0,dwSize);  
	assert(pBuffer);

	shared_queue.m_pHeader=(SHARED_QUEUE_HEADER*)pBuffer;
	shared_queue.m_pQueue=(DWORD*)(pBuffer+sizeof(SHARED_QUEUE_HEADER));
	shared_queue.m_hMutex=CreateMutex(NULL,FALSE,GUARD_MUTEX_NAME);
	assert(shared_queue.m_hMutex);

	if(!bConnect) // if shared queue not created yet
	{
		BOOL bRes=InitializeSharedQueue(&shared_queue);
		assert(bRes);
		_tprintf(_T("queue was created\n"));
	}
	else
	{
		_tprintf(_T("attached to queue\n"));
	}

	// obtain reader semaphore; initialize with current shared_queue data count
	HANDLE hReaderSem=CreateSemaphore(NULL,shared_queue.m_pHeader->m_dwCount,QUEUE_MAX_COUNT,READER_SEMAPHORE);
	assert(hReaderSem!=NULL);
	// obtain writer semaphore; initialize with current available space in queue
	HANDLE hWriterSem=CreateSemaphore(NULL,(QUEUE_MAX_COUNT-shared_queue.m_pHeader->m_dwCount),QUEUE_MAX_COUNT,WRITER_SEMAPHORE);
	assert(hWriterSem!=NULL);

	do // main loop
	{
		// wait until there is data in the queue:
		// call READER semaphore and update its use
		// when shared_queue is empty, will have to wait for writer to fill it
		DWORD dwRes=WaitForSingleObject(hReaderSem,ALMOST_INFINITE);
		assert(dwRes==WAIT_OBJECT_0);

		// pop value from shared_queue
		QUEUE_COMMAND qc;
		qc.m_cc=CC_POP;
		BOOL bRes=ProcessQueueCommand(&shared_queue, &qc);
		assert(bRes);

		// release writer semaphore (space available in the queue)
		bRes=ReleaseSemaphore(hWriterSem, 1, NULL);
		assert(bRes);

		// print popped value and sleep
		DWORD dwValue=qc.m_dwValue;
		_tprintf(_T("%d\n"),dwValue);
		Sleep(dwValue);

		// check if should exit
		if (!dwValue) break;
	
	}while(1);

	CloseHandle(shared_queue.m_hMutex);
	UnmapViewOfFile(pBuffer);
	CloseHandle(hMapFile);
	CloseHandle(hReaderSem);
	CloseHandle(hWriterSem);

	return 0;
}