// lifo_queue.cpp : Defines the exported functions for the DLL application.
//

#include "stdafx.h"
#include "../mmf_queue/mmf_queue.h"

extern "C"
{
	BOOL __declspec(dllimport) sharedCreateQueue(SHARED_QUEUE *pSQ);
	BOOL __declspec(dllimport) sharedDeleteQueue(SHARED_QUEUE *pSQ);
	BOOL __declspec(dllexport) pfnCreateQueue(SHARED_QUEUE *pSQ);
	BOOL __declspec(dllexport) pfnDeleteQueue(SHARED_QUEUE *pSQ);
	BOOL __declspec(dllexport) pfnPushElement(SHARED_QUEUE *pSQ, DWORD dwValue);
	BOOL __declspec(dllexport) pfnPopElement(SHARED_QUEUE *pSQ, DWORD *dwValue);
}

BOOL pfnCreateQueue(SHARED_QUEUE *pSQ){
	return sharedCreateQueue(pSQ);
}

BOOL pfnDeleteQueue(SHARED_QUEUE *pSQ){
	return sharedDeleteQueue(pSQ);
}

// LIFO implementation for pushing element to queue
BOOL pfnPushElement(SHARED_QUEUE *pSQ, DWORD dwValue){
	// wait until there's space in the queue
	DWORD dwRes=WaitForSingleObject(pSQ->m_hWriteSem, ALMOST_INFINITE);
	assert(dwRes==WAIT_OBJECT_0);

	// get queue read-write mutex
	dwRes=WaitForSingleObject(pSQ->m_hMutex,ALMOST_INFINITE);
	assert(dwRes==WAIT_OBJECT_0);

	// if queue is full, return false (should be taken care of by the writer semaphore)
	if (QUEUE_MAX_COUNT==pSQ->m_pHeader->m_dwCount) {
		BOOL bRes=ReleaseMutex(pSQ->m_hMutex);
		assert(bRes);
		return FALSE;
	}
	
	// else add data to queue
	pSQ->m_pQueue[pSQ->m_pHeader->m_dwCount++]=dwValue;
	
	// release queue read-write mutex
	BOOL bRes=ReleaseMutex(pSQ->m_hMutex);
	assert(bRes);

	// release reader semaphore
	bRes=ReleaseSemaphore(pSQ->m_hReadSem,1,NULL);
	assert(bRes);

	return TRUE;
}

// LIFO implementation for popping element from queue
BOOL pfnPopElement(SHARED_QUEUE *pSQ, DWORD *dwValue){
	// wait until there's data in the queue
	DWORD dwRes=WaitForSingleObject(pSQ->m_hReadSem, ALMOST_INFINITE);
	assert(dwRes==WAIT_OBJECT_0);

	// get queue read-write mutex
	dwRes=WaitForSingleObject(pSQ->m_hMutex,ALMOST_INFINITE);
	assert(dwRes==WAIT_OBJECT_0);

	// if queue is empty, return false (should be taken care of by reader semaphore)
	if (!(pSQ->m_pHeader->m_dwCount)) {
		BOOL bRes=ReleaseMutex(pSQ->m_hMutex);
		assert(bRes);
		return FALSE;
	}
	
	// else read data from queue and place in dwValue
	(*dwValue)=pSQ->m_pQueue[pSQ->m_pHeader->m_dwCount-1];
	pSQ->m_pHeader->m_dwCount--;
	
	// release queue read-write mutex
	BOOL bRes=ReleaseMutex(pSQ->m_hMutex);
	assert(bRes);

	// release writer semaphore
	bRes=ReleaseSemaphore(pSQ->m_hWriteSem,1,NULL);
	assert(bRes);

	return TRUE;
}