// mmf_queue.cpp : Defines the exported functions for the DLL application.
//

#include "stdafx.h"
#include "mmf_queue.h"

extern "C"
{
	BOOL __declspec(dllexport) sharedCreateQueue(SHARED_QUEUE *pSQ);
	BOOL __declspec(dllexport) sharedDeleteQueue(SHARED_QUEUE *pSQ);
}

// initialize queue (private)
BOOL InitializeSharedQueue(SHARED_QUEUE* pSQ)
{
	DWORD dwRes=WaitForSingleObject(pSQ->m_hMutex,ALMOST_INFINITE);
	assert(dwRes==WAIT_OBJECT_0);

	ZeroMemory(pSQ->m_pHeader,sizeof(SHARED_QUEUE_HEADER));

	ReleaseMutex(pSQ->m_hMutex);

	return TRUE;
}

// create queue (public)
BOOL sharedCreateQueue(SHARED_QUEUE* pSQ)
{
	// create shared queue in memory and initialize handles
	DWORD dwSize=(QUEUE_MAX_COUNT * sizeof(DWORD))+ sizeof(SHARED_QUEUE_HEADER);
	SHARED_QUEUE shared_queue;

	shared_queue.m_hMapFile = CreateFileMapping(INVALID_HANDLE_VALUE,NULL,PAGE_READWRITE,0,dwSize,MMF_NAME);
	assert(shared_queue.m_hMapFile);

	BOOL bConnect=(GetLastError()==ERROR_ALREADY_EXISTS);

	shared_queue.m_pBuffer=(BYTE*)MapViewOfFile(shared_queue.m_hMapFile,FILE_MAP_WRITE|FILE_MAP_WRITE,0,0,dwSize);  
	assert(shared_queue.m_pBuffer);

	shared_queue.m_pHeader=(SHARED_QUEUE_HEADER*)shared_queue.m_pBuffer;
	shared_queue.m_pQueue=(DWORD*)(shared_queue.m_pBuffer+sizeof(SHARED_QUEUE_HEADER));

	// create read-write mutex, reader and writer semaphores
	shared_queue.m_hMutex=CreateMutex(NULL,FALSE,GUARD_MUTEX_NAME);
	assert(shared_queue.m_hMutex);
	shared_queue.m_hReadSem=CreateSemaphore(NULL,0,QUEUE_MAX_COUNT-1,READER_SEM_NAME);
	assert(shared_queue.m_hReadSem);
	shared_queue.m_hWriteSem=CreateSemaphore(NULL,QUEUE_MAX_COUNT-1,QUEUE_MAX_COUNT-1,WRITER_SEM_NAME);
	assert(shared_queue.m_hWriteSem);

	if(!bConnect) // if shared queue not created yet
	{
		BOOL bRes=InitializeSharedQueue(&shared_queue);
		assert(bRes);
		_tprintf(_T("queue was created\n"));
	}
	else
	{
		_tprintf(_T("attached to queue\n"));
	}

	(*pSQ) = shared_queue;
	return TRUE;
}

// delete queue (public)
BOOL sharedDeleteQueue(SHARED_QUEUE* pSQ){
	// close all handles
	CloseHandle(pSQ->m_hMutex);
	CloseHandle(pSQ->m_hReadSem);
	CloseHandle(pSQ->m_hWriteSem);
	UnmapViewOfFile(pSQ->m_pBuffer);
	CloseHandle(pSQ->m_hMapFile);

	return TRUE;
}